/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


function Model(config)
{
    for (var [name, entry] in Iterator(config))
    {
        var value = (("value" in entry) ? entry.value : null);
        var depends = (("depends" in entry) ? entry.depends : "");
        var equals = (("equals" in entry) ? entry.equals : undefined);
        this.addProperty(name, value, depends, equals);
    }
}

Model.prototype =
{
    addProperty: function(name, value, depends, equals)
    {
        var _name = "_" + name;
        var entry = {};
        
        entry.value = entry.initialValue = value;
        entry.equals = equals;
        entry.children = [];
        entry.watchers = [];
        
        var getter = function() { return this[_name].value; };
        var setter = function(val)
        {
            var old = entry.value;
            
            var equals = (entry.equals ? entry.equals(val, old) : (val === old));
            if (! equals)
            {
                entry.value = val;
                for each (var child in entry.children)
                    this[child] = this["_" + child].initialValue;
                for each (var watcher in entry.watchers)
                    watcher(name, val, old);
            }
            
            return val;
        };
        
        this[_name] = entry;
        this.__defineGetter__(name, getter);
        this.__defineSetter__(name, setter);
        
        if (depends)
            for each (var dep in depends.split(/[,\s]+/))
                if (dep == name)
                    throw new Error("property '" + name + "' cannot depend on itself");
                else if (! (dep in this))
                    throw new Error("property '" + name + "' depends on property '" + dep + "' but it doesn't exist in model");
                else
                    this["_" + dep].children.push(name);
    },
    
    addPropertyWatcher: function(watcher, properties)
    {
        for each (var name in properties)
            this["_" + name].watchers.push(watcher);
    }
}
